package slimeknights.tconstruct.library.modifiers.modules.behavior;

import net.minecraft.nbt.Tag;
import net.minecraft.network.chat.Component;
import net.minecraft.resources.ResourceLocation;
import slimeknights.mantle.data.loadable.primitive.IntLoadable;
import slimeknights.mantle.data.loadable.record.RecordLoadable;
import slimeknights.mantle.data.registry.GenericLoaderRegistry.IGenericLoader;
import slimeknights.tconstruct.library.modifiers.Modifier;
import slimeknights.tconstruct.library.modifiers.ModifierHook;
import slimeknights.tconstruct.library.modifiers.TinkerHooks;
import slimeknights.tconstruct.library.modifiers.hook.behavior.EffectiveLevelModifierHook;
import slimeknights.tconstruct.library.modifiers.hook.build.ModifierRemovalHook;
import slimeknights.tconstruct.library.modifiers.hook.display.DisplayNameModifierHook;
import slimeknights.tconstruct.library.modifiers.impl.IncrementalModifier;
import slimeknights.tconstruct.library.modifiers.modules.ModifierModule;
import slimeknights.tconstruct.library.modifiers.util.ModuleWithKey;
import slimeknights.tconstruct.library.recipe.modifiers.ModifierRecipeLookup;
import slimeknights.tconstruct.library.tools.nbt.INamespacedNBTView;
import slimeknights.tconstruct.library.tools.nbt.IToolContext;
import slimeknights.tconstruct.library.tools.nbt.IToolStackView;

import javax.annotation.Nullable;
import java.util.List;

/**
 * Module that makes a modifier incremental.
 * @param key             Persistent data key containing the incremental data. If null, uses the modifier ID.
 *                        Presently, changing this makes it incompatible with the incremental modifier recipe, this is added for future proofing.
 * @param neededPerLevel  If zero, modifier is incremental with no max set and will fetch if from the recipe.
 *                        If greater than zero, modifier will have a fixed max.
 */
public record IncrementalModule(@Nullable ResourceLocation key, int neededPerLevel) implements EffectiveLevelModifierHook, DisplayNameModifierHook, ModifierRemovalHook, ModifierModule, ModuleWithKey {
  private static final List<ModifierHook<?>> DEFAULT_HOOKS = List.of(TinkerHooks.EFFECTIVE_LEVEL, TinkerHooks.DISPLAY_NAME, TinkerHooks.REMOVE);
  public static RecordLoadable<IncrementalModule> LOADER = RecordLoadable.create(ModuleWithKey.FIELD, IntLoadable.FROM_ZERO.defaultField("needed_per_level", 0, IncrementalModule::neededPerLevel), IncrementalModule::new);

  /** Recipe controlled incremental modifier with no extra settings */
  public static IncrementalModule RECIPE_CONTROLLED = new IncrementalModule(null, 0);

  /** Just saving a bit of memory as most things use the recipe controlled constant */
  public static IncrementalModule create(@Nullable ResourceLocation key, int neededPerLevel) {
    if (key == null && neededPerLevel == 0) {
      return RECIPE_CONTROLLED;
    }
    return new IncrementalModule(key, neededPerLevel);
  }

  /** Gets the number needed per level */
  private int getNeededPerLevel(ResourceLocation key) {
    if (neededPerLevel > 0) {
      return neededPerLevel;
    }
    return ModifierRecipeLookup.getNeededPerLevel(key);
  }

  /** Gets the amount presently on the tool */
  private int getAmount(IToolContext tool, ResourceLocation key, int neededPerLevel) {
    INamespacedNBTView data = tool.getPersistentData();
    if (data.contains(key, Tag.TAG_ANY_NUMERIC)) {
      return data.getInt(key);
    }
    return neededPerLevel;
  }

  @Override
  public Component getDisplayName(IToolStackView tool, Modifier modifier, int level, Component name) {
    ResourceLocation key = getKey(modifier);
    int neededPerLevel = getNeededPerLevel(key);
    if (neededPerLevel > 0) {
      return IncrementalModifier.addAmountToName(getAmount(tool, key, neededPerLevel), neededPerLevel, name);
    }
    return name;
  }

  @Override
  public float getEffectiveLevel(IToolContext tool, Modifier modifier, float level) {
    if (level <= 0) {
      return 0;
    }
    ResourceLocation key = getKey(modifier);
    int neededPerLevel = getNeededPerLevel(key);
    if (neededPerLevel > 0) {
      return IncrementalModifier.scaleLevel(level, getAmount(tool, key, neededPerLevel), neededPerLevel);
    }
    return level;
  }

  @Nullable
  @Override
  public Component onRemoved(IToolStackView tool, Modifier modifier) {
    tool.getPersistentData().remove(modifier.getId());
    return null;
  }

  @Override
  public List<ModifierHook<?>> getDefaultHooks() {
    return DEFAULT_HOOKS;
  }

  @Override
  public IGenericLoader<? extends ModifierModule> getLoader() {
    return LOADER;
  }
}
